/**************************************************************************
***    
***    Copyright (c) 1995-1999 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA, 
***    and by the Regents of the University of California.    
***    All rights reserved.   
***
***    No portion of this package may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/

// Created by Mike Oliver on 3 april 1999
//////////////////////////////////////////////////////////////////////
//
// zTolPart.cxx: implementation of the ZTolPart class.
//
//////////////////////////////////////////////////////////////////////

#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

#include "zTolPart.h"

#include <HMetis/hmetisPart.h>
#include <FMPart/fmPart.h>
#include <SmallPart/bbPart.h>
#include <PartLegality/1balanceChk.h>


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ZTolPart::ZTolPart(PartitioningProblem &problem,
                   Parameters params)
	:_problem(problem), _params(params),
	 _hgraph(problem.getHGraph()), _eval(problem),
	 _nodesVisited(_hgraph.getNumNodes()),
	 _edgesVisited(_hgraph.getNumEdges()),
	 _mapBack(_hgraph.getNumNodes(),UINT_MAX),
	 _mapBackCutline(_hgraph.getNumNodes(),UINT_MAX),
	 _isCutline(_hgraph.getNumNodes())
{
    PartitioningBuffer &buf = _problem.getSolnBuffers();

    abkfatal(problem.getNumPartitions()==2,"ZTolPart is only for 2-way "
                                           "partitioning.");
    abkfatal(problem.getMaxCapacities()[0] == problem.getMinCapacities()[0]
          && problem.getMaxCapacities()[1] == problem.getMinCapacities()[1],
          "ZTolPart is only for zero-tolerance partitioning problems");

    _initMovables();
        
    HMetisParams hmParams;
    hmParams.startsPerSolution = params.numHMetisStarts;
    HMetisPartitioner hParter(problem,hmParams);

    _eval.resetTo(buf[_problem.getBestSolnNum()]);

    cout << "Cut after " << params.numHMetisStarts << " HM starts is " <<
        _eval.getTotalCost() << endl;

    cout << "Violations after HMetis are " 
        << _problem.getViolations()[_problem.getBestSolnNum()] << endl;

    FMPartitioner::Parameters fmParams;
    FMPartitioner    fmParter(problem,fmParams);

    _eval.resetTo(buf[_problem.getBestSolnNum()]);

    cout << "Cut after FM is " <<
        _eval.getTotalCost() << endl;

    cout << "Violations after FM are " 
        << _problem.getViolations()[_problem.getBestSolnNum()] << endl;


    BBPart::Parameters bbParams; bbParams.verb=Verbosity("0_0_0");
    for (unsigned m=0;m<_params.numBBPartCycles;m++)
        {
        unsigned patchCount=0;
        itHGFEdgeGlobal iE=_hgraph.edgesBegin();
        for (;iE!=_hgraph.edgesEnd();iE++)
            {
            HGFEdge const &edge=*(*iE);
            unsigned edgeIdx = edge.getIndex();
            if (_eval.computeCostOfOneNet(edgeIdx)>0
                && edge.getDegree() < _params.maxComponentSize / 2)
                {
                _patch.clear();
                patchCount++;
                //_findPatchFromEdge(edgeIdx); 
                _findCutline();
                _findPatchBFS(edgeIdx);
                //cout << "Trying patch of size" << _patch.size() << endl;
                //cout << "Patch indices are: " << endl;
                //for (unsigned l=0;l<_patch.size();l++)
                //    cout << " " << _patch[l];
                _refinePatch();
                }
            }
        _eval.resetTo(buf[_problem.getBestSolnNum()]);
        cout << "Refined " << patchCount << " patches, cut = " 
             << _eval.getTotalCost() << endl;
        }


    _finalize();
}

ZTolPart::~ZTolPart()
    {
    
    }

void ZTolPart::_initMovables()
    {
    const Partitioning& fixedConstr=_problem.getFixedConstr();
    PartitionIds part0, part1, part01;
    part0.setToPart(0); part1.setToPart(1); part01.setToAll(2);
    unsigned nNodes=_hgraph.getNumNodes();
    for(unsigned k=0; k!=nNodes; k++)
        {
        if ( fixedConstr[k].isEmpty() || fixedConstr[k]==part01 ) 
            {
            _movables.push_back(k);
            }
        }

    for (k=0;k<_movables.size();k++)
        {
        _mapBack[_movables[k]]=k;
        }
    }

void ZTolPart::_finalize()
    {
    PartitioningBuffer& buf=_problem.getSolnBuffers();
    unsigned solNo = _problem.getBestSolnNum();
    
    vector<double>& viols=const_cast<vector<double>&>(_problem.getViolations());
    // the latter won't be set
    vector<double>& costs=const_cast<vector<double>&>(_problem.getCosts());

    _eval.resetTo(buf[solNo]);

    costs[solNo] = _eval.getTotalCost();

    SingleBalanceLegality leg(_problem,buf[solNo]);

    viols[solNo]  = leg.getMaxViolation();

    
    }

void ZTolPart::_findCutline()
    {
    PartitioningBuffer &buf = _problem.getSolnBuffers();
    unsigned soln = _problem.getBestSolnNum();
    Partitioning &part = buf[soln];
    _eval.resetTo(part);
    itHGFEdgeGlobal iE=_hgraph.edgesBegin();
    for (;iE!=_hgraph.edgesEnd();iE++)
        {
        HGFEdge const &edge = *(*iE);
        if (_eval.computeCostOfOneNet(edge.getIndex())>0 &&
            edge.getDegree() < _params.netIgnoreSize)
            {
            itHGFNodeLocal iN=edge.nodesBegin();
            for (;iN!=edge.nodesEnd();iN++)
                {
                unsigned nodeIdx=(*iN)->getIndex();
                if (!_nodesVisited.isBitSet(nodeIdx) &&
                    _mapBack[nodeIdx] != UINT_MAX)
                    {
                    _nodesVisited.setBit(nodeIdx);
                    _cutline.push_back(nodeIdx);
                    _isCutline[nodeIdx]=true;
                    }

                }


            }
        }
    for (unsigned k=0;k<_cutline.size();k++)
        {
        _mapBackCutline[_cutline[k]]=k;
        }
    _nodesVisited.clear();
    }

void ZTolPart::_findComponents()
    {
    _findCutline();

    for (unsigned k=0;k<_cutline.size();k++)
        {
        if (!_nodesVisited.isBitSet(_cutline[k]))
            {
            _components.push_back(vector<unsigned>());
            _expandNode(_cutline[k]);
            }
        }
    }

void ZTolPart::_expandNode(unsigned nodeIdx)
    {

    _nodesVisited.setBit(nodeIdx);
    _components.back().push_back(nodeIdx);

    if (_components.back().size() >= _params.maxComponentSize)
        {
        _components.push_back(vector<unsigned>());  //start new patch because
                                                    //old one too big
        }

    HGFNode const &node = _hgraph.getNodeByIdx(nodeIdx);

    itHGFEdgeLocal iE=node.edgesBegin();

    for (;iE!=node.edgesEnd();iE++)
        {
        HGFEdge const &edge = *(*iE);
        unsigned edgeIdx=edge.getIndex();
        if (!_edgesVisited.isBitSet(edgeIdx) &&
            edge.getDegree() < _params.netIgnoreSize)
            {
            _expandEdge(edgeIdx);
            }
        }

    }

void ZTolPart::_expandEdge(unsigned edgeIdx)
    {
    _edgesVisited.setBit(edgeIdx);

    HGFEdge const &edge = _hgraph.getEdgeByIdx(edgeIdx);

    itHGFNodeLocal iN=edge.nodesBegin();

    for (;iN!=edge.nodesEnd();iN++)
        {
        unsigned nodeIdx = (*iN)->getIndex();
        if (_mapBackCutline[nodeIdx] != UINT_MAX // node on cutline
            || _rrd < _params.nonCutlineIncludeProb )
            {
            if (!_nodesVisited.isBitSet(nodeIdx))
                {
                _expandNode(nodeIdx);
                }
            }
        }
    }


void ZTolPart::_refineComponent(unsigned compIdx)
    {
    PartitioningBuffer &buf = _problem.getSolnBuffers();
    unsigned soln = _problem.getBestSolnNum();
    Partitioning &part = buf[soln];
    double area0=0.0,area1=0.0;
    vector<unsigned> const &comp = _components[compIdx];

    for (unsigned k=0;k<comp.size();k++)
        {
        unsigned nodeIdx=comp[k];
        abkassert(!part[nodeIdx].isInPart(0) || !part[nodeIdx].isInPart(1),
                  "Incomplete solution in _refineComponent");
        if (part[nodeIdx].isInPart(0))
            {
            area0 += _hgraph.getNodeByIdx(nodeIdx).getWeight();
            }
        else
            {
            area1 += _hgraph.getNodeByIdx(nodeIdx).getWeight();
            }
        }

    BBPart::Parameters bbParams;
    bbParams.verb = _params.verb;

    BBPart(_hgraph,part,comp,area0,area1,bbParams);

    }

void ZTolPart::_refinePatch()
    {
    PartitioningBuffer &buf = _problem.getSolnBuffers();
    unsigned soln = _problem.getBestSolnNum();
    Partitioning &part = buf[soln];
    double area0=0.0,area1=0.0;

    for (unsigned k=0;k<_patch.size();k++)
        {
        unsigned nodeIdx=_patch[k];
        abkassert(!part[nodeIdx].isInPart(0) || !part[nodeIdx].isInPart(1),
                  "Incomplete solution in _refineComponent");
        if (part[nodeIdx].isInPart(0))
            {
            area0 += _hgraph.getNodeByIdx(nodeIdx).getWeight();
            }
        else
            {
            area1 += _hgraph.getNodeByIdx(nodeIdx).getWeight();
            }
        }

    BBPart::Parameters bbParams;
    bbParams.verb = Verbosity("0_0_0");

    BBPart(_hgraph,part,_patch,area0,area1,bbParams);

    }
void ZTolPart::_findPatchFromEdge(unsigned edgeIdx)
    {
    _depthCounter=0;
    _edgesVisited.clear();
    _nodesVisited.clear();
    _expandEdgeOnePatch(edgeIdx);
    }


void ZTolPart::_expandNodeOnePatch(unsigned nodeIdx)
    {
    if (_nodesVisited.isBitSet(nodeIdx))
        return;
    _nodesVisited.setBit(nodeIdx);

    ++_depthCounter;
    _patch.push_back(nodeIdx);

    if (_patch.size() >= _params.maxComponentSize)
        {
        return;
        }

    HGFNode const &node = _hgraph.getNodeByIdx(nodeIdx);

    itHGFEdgeLocal iE=node.edgesBegin();

    for (;iE!=node.edgesEnd();iE++)
        {
        HGFEdge const &edge = *(*iE);
        unsigned edgeIdx=edge.getIndex();
        if (!_edgesVisited.isBitSet(edgeIdx) &&
            edge.getDegree() < _params.netIgnoreSize)
            {
            _expandEdgeOnePatch(edgeIdx);
            if (_patch.size() >= _params.maxComponentSize)
                {
                return;
                }
            }
        }

    --_depthCounter;
    }
void ZTolPart::_expandEdgeOnePatch(unsigned edgeIdx)
    {
    if (_depthCounter>=_params.depthLimit)
        return;
    _edgesVisited.setBit(edgeIdx);

    HGFEdge const &edge = _hgraph.getEdgeByIdx(edgeIdx);

    itHGFNodeLocal iN=edge.nodesBegin();

    for (;iN!=edge.nodesEnd();iN++)
        {
        unsigned nodeIdx = (*iN)->getIndex();
        if (!_nodesVisited.isBitSet(nodeIdx))
            {
            _expandNodeOnePatch(nodeIdx);
            if (_patch.size() >= _params.maxComponentSize)
                {
                return;
                }
            }
        }
    }

void ZTolPart::_findPatchBFS(unsigned edgeIdx)
    {
    _nodesVisited.clear();
    RandomUnsigned ru(0,1);  // the bounds are meaningless; we're using
                             // operator[] in random_shuffle
    
    HGFEdge const &startEdge = _hgraph.getEdgeByIdx(edgeIdx);
    itHGFNodeLocal iN = startEdge.nodesBegin();
    for (;iN!=startEdge.nodesEnd();iN++)
        {
        HGFNode const & node=*(*iN);
        _patch.push_back(node.getIndex());
        _nodesVisited.setBit(node.getIndex());
        }
    vector <unsigned> curNeighb=_patch,toProcess;
    while (!curNeighb.empty())
        {
        toProcess=curNeighb;
        std::random_shuffle(toProcess.begin(),toProcess.end(),ru);
        unsigned neighbSize=curNeighb.size();
        curNeighb.clear();

        for (unsigned k=0;k<neighbSize;k++)
            {
            unsigned idx=toProcess.back();
            toProcess.pop_back();
            HGFNode const &node=_hgraph.getNodeByIdx(idx);
            itHGFEdgeLocal iE = node.edgesBegin();
            for (;iE!=node.edgesEnd();iE++)
                {
                HGFEdge const &edge = *(*iE);
                itHGFNodeLocal iN2=edge.nodesBegin();
                for (;iN2!=edge.nodesEnd();iN2++)
                    {
                    unsigned nodeIdx = (*iN2)->getIndex();
                    if (nodeIdx==idx) continue;
                    if (_nodesVisited.isBitSet(nodeIdx)) continue;
                    _nodesVisited.setBit(nodeIdx);
                    if (_isCutline[nodeIdx])
                        curNeighb.push_back(nodeIdx);
                    _patch.push_back(nodeIdx);
                    if (_patch.size() >= _params.maxComponentSize)
                        return;
                    }
                }
            }

        }
    }


